"""
    ...
"""
from qt import *
import operator
import config
import kdeui
import random
from util import find, exists
import datetime
class MyPoint:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        
class FlashingLabel(QLabel):
    def __init__(self, text, parent):
        QLabel.__init__(self, text, parent)

        # BEGIN works with plastik but not keramik
        #red = pal.text().red()*low + pal.highlightedText().red()*high
        #green = pal.text().green()*low + pal.highlightedText().green()*high
        #blue = pal.text().blue()*low + pal.highlightedText().blue()*high
        # END

    
    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(90)
    def stopFlashing(self, resetColor):
        self.killTimers()
        if resetColor:
            self.flashingState = FlashingStateNo()
            self.setPaletteForegroundColor(self.palette().active().text())
        else:
            self.flashingState = FlashingStateYesStill()
            self.setPaletteForegroundColor(self.highlightColor)
            
    def timerEvent(self, ev):
        pal = self.palette().active()
        # BEGIN compute new text color for the items
        col = QColor() #(63, 143, 91)
        col.setHsv( (self.flashingState.remainingFlashesBeforeStop*40) % 255
                    , 133
                    ,(self.flashingState.remainingFlashesBeforeStop*50) % 255)
        # END
        self.setPaletteForegroundColor(col)
        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)


class FlashingTextEdit(kdeui.KTextEdit):
    def __init__(self, parent):
        kdeui.KTextEdit.__init__(self, parent)
        self.originalBackgroundColor = self.paletteBackgroundColor()
        self.originalForegroundColor = self.paletteForegroundColor()
        self.timerId = None
    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(
            howManyFlashes, resetOriginalColorsWhenFinished)
        self.timerId =self.startTimer(150)
    def stopFlashing(self, resetColor):
        if self.timerId:
            self.killTimer(self.timerId)
            self.timerId = None
        if resetColor:
            self.flashingState = FlashingStateNo()
            self.setPaletteBackgroundColor(self.originalBackgroundColor)
            self.setPaletteForegroundColor(self.originalForegroundColor)
        else:
            self.flashingState = FlashingStateYesStill()
            self.setPaletteBackgroundColor(self.palette().active().highlight())
            self.setPaletteForegroundColor(self.palette().active().highlightedText())
    # FlashingTextEdit
    def timerEvent(self, ev):
        if ev.timerId() == self.timerId:
            mod = self.flashingState.remainingFlashesBeforeStop % 2
            if mod == 0:
                self.setPaletteBackgroundColor(self.palette().active().light())
            else:
                self.setPaletteBackgroundColor(self.palette().active().mid())

            self.flashingState.remainingFlashesBeforeStop -= 1
            if self.flashingState.remainingFlashesBeforeStop <= 0:
                self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)
    


class AbstractFlashingButton:
    def init(self):
        self.originalBackgroundColor = self.paletteBackgroundColor()
        self.originalForegroundColor = self.paletteForegroundColor()
        self.button_brush = self.palette().active().brush(QColorGroup.Button)
        self.button_text_brush = self.palette().active().brush(QColorGroup.ButtonText)
        background_brush = self.palette().active().brush(QColorGroup.Background)
        self.is_pixmapped_style = background_brush.pixmap() != None
    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(
            howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(150)
    def stopFlashing(self, resetColor = True):
        self.killTimers()
        if resetColor:
            self.flashingState = FlashingStateNo()
            if not self.is_pixmapped_style:
                self.setPaletteBackgroundColor(self.originalBackgroundColor)
                self.setPaletteForegroundColor(self.originalForegroundColor)
            else:
                self.palette().active().setBrush(QColorGroup.Button
                                           , self.button_brush)
                self.palette().active().setBrush(QColorGroup.ButtonText
                                           , self.button_text_brush)
        else:
            self.flashingState = FlashingStateYesStill()
            if not self.is_pixmapped_style:
                self.setPaletteBackgroundColor(self.palette().active().highlight())
                self.setPaletteForegroundColor(self.palette().active().highlightedText())
            else:
                self.palette().active().setBrush(QColorGroup.Button
                                                 , QBrush(self.palette().active().highlight()
                                                          , Qt.SolidPattern))
                self.palette().active().setBrush(QColorGroup.ButtonText
                                                 , QBrush(self.palette().active().highlightedText()
                                                          , Qt.SolidPattern))
        self.repaint()
    # FlashingButton
    def timerEvent(self, ev):
        mod = self.flashingState.remainingFlashesBeforeStop % 2
        if not self.is_pixmapped_style:
            if mod == 0:
                self.setPaletteBackgroundColor(self.palette().active().mid())
            else:
                self.setPaletteBackgroundColor(self.palette().active().light())
        else:
            if mod == 0:
                self.palette().active().setBrush(QColorGroup.Button
                                                 , QBrush(self.palette().active().highlight()
                                                          , Qt.SolidPattern))
            else:
                self.palette().active().setBrush(QColorGroup.Button
                                                 , self.button_brush)

        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)
        self.repaint()

class FlashingPushButton(QPushButton, AbstractFlashingButton):
    def __init__(self, icon, text, parent):
        QPushButton.__init__(self, icon, text, parent)
        AbstractFlashingButton.init(self)


class FlashingToolButton(QToolButton):
    def __init__(self, icon, text, group_text, callback, parent):
        QToolButton.__init__(self, icon, text, group_text, callback, parent)

    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(
            howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(150)
    def stopFlashing(self, resetColor = True):
        self.killTimers()
        self.flashingState = FlashingStateNo()
        self.setAutoRaise(True)
        #        self.repaint()
    # FlashingToolButton
    def timerEvent(self, ev):
        mod = self.flashingState.remainingFlashesBeforeStop % 2
        if mod == 0:
            self.setAutoRaise(True)
        else:
            self.setAutoRaise(False)

        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)
        #        self.repaint()

class ListArea(QWidget):
    def __init__(self, parentWidget, parentList):
        QWidget.__init__(self, parentWidget)
        self.parentList = parentList
        self.setMouseTracking(True)
        self.setAcceptDrops(True)
        # BEGIN can't do this because if I do, pixmapped styles loose
        # the pixmap; when I call fill() in paint(), the ListArea is
        # filled with black.
        #self.setBackgroundMode(Qt.NoBackground)
        # END

        #self.paintcount = 0 # @@ debug
    def dragEnterEvent(self, event):
        event.accept( True)
        self.drag_timer_id = None
    def dropEvent(self, ev):
        ev.acceptAction(False)
        #         it = self.parentList.itemAtPosition(ev.pos().y())
        #         data = self.parentList.items[it].data
        #         self.emit(PYSIGNAL("dropped_on_item")
        #                   , (ev, data))
        #         print "dropped_on_item " + str(data)
    def dragMoveEvent(self, ev):
        print "dragMoveEvent"
        # BEGIN remember mouse pos
        currentPosx = ev.pos().x()
        currentPosy = ev.pos().y()
        current_pos = MyPoint(currentPosx, currentPosy)
        if self.parentList.lastMousePos != None:
            ys = [ self.parentList.lastMousePos.y, ev.pos().y()]
            ys.sort()
            self.paint(current_pos , ys[0], ys[1])
        else:
            self.paint(current_pos , None, None)
        self.parentList.lastMousePos = current_pos
        # END
        if self.drag_timer_id != None:
            self.killTimer(self.drag_timer_id)
        self.drag_timer_id = self.startTimer(1500)
        ev.accept()
    def timerEvent(self, ev):
        print "timerEvent"
        self.killTimer(self.drag_timer_id)
        self.drag_timer_id = None
        if self.parentList.lastMousePos:
            item_idx = self.parentList.itemAtPosition(self.parentList.lastMousePos.y)
            item = self.parentList.items[item_idx]
            if isinstance(item, TextItem):
                wid = item.data
                self.parentList.emit(PYSIGNAL("drag_activate"), (wid,))
    def wheelEvent(self, ev):
        movedUp = ev.delta() > 0
        self.parentList.centeringState = None
        if movedUp:
            self.parentList.scrollUpLineWithoutUpdating()
        else:
            self.parentList.scrollDownLineWithoutUpdating()
        #print "wheelEvent: calling make_changes_visible"
        self.parentList.make_changes_visible()
        
    def paintEvent(self, ev):
        #print "ListArea.paintEvent. pos = " + str(self.parentList.lastMousePos)
        self.paint(self.parentList.lastMousePos
                   , ev.rect().top()
                   , ev.rect().bottom())
    #ListArea.paint
    def paint(self, mousePos, invalidated_y_top, invalidated_y_bottom):
        #print "paint. Area = " + str(self) + " ; count = " + str(self.paintcount)
        #self.paintcount +=1
        def top_y_of_item(i):
            return reduce(operator.add
                          ,[x.height for x in self.parentList.items[self.parentList.scrollingOffset : i]]
                          , 0)
        def bottom_y_of_item(i):
            return reduce(operator.add
                          , [ x.height for x in self.parentList.items[self.parentList.scrollingOffset : i+1]]
                          , 0)
        def min_list(l):
            m = l[0]
            for i in l:
                if i < m:
                    m = i
            return m
        def max_list(l):
            m = l[0]
            for i in l:
                if i > m:
                    m = i
            return m
        assert (invalidated_y_top == None and invalidated_y_bottom == None
                or
                invalidated_y_top <= invalidated_y_bottom)
        max_items_visible_at_once = int(self.height() / config.gMinItemHt) + 1
        end_index = min( len(self.parentList.items)
                         , self.parentList.scrollingOffset + max_items_visible_at_once)
        if invalidated_y_top:
            items_to_draw2 = [ i for i in range(self.parentList.scrollingOffset
                                                , end_index)
                               if (i == self.parentList.itemAtPosition(invalidated_y_top)
                                   or
                                   i == self.parentList.itemAtPosition(invalidated_y_bottom))]
        else:
            items_to_draw2 = []
        if invalidated_y_top == None or items_to_draw2 == []:
            items_to_draw = range(0, len(self.parentList.items))
            pixmap_blitting_y = 0
            pixmap_height = self.height()
        else:
            pixmap_blitting_y = min_list ( [top_y_of_item(i) for i in items_to_draw2])
            pixmap_height = max_list( [ bottom_y_of_item(i) for i in items_to_draw2]) - pixmap_blitting_y
            items_to_draw = range(items_to_draw2[0], items_to_draw2[-1]+1)
        
        pix = QPixmap(self.width(), pixmap_height)
        assert pix != None
        painter = QPainter(pix) # begin painting on the pixmap
        assert painter != None

        # BEGIN If I set the background mode to NoBackground, this
        # fills in black , instead of using the pixmap. OTOH, if I do
        # NOT set the background mode to NoBackground, this should not
        # be needed, instead it is needed. Why?
        #
        # BUG: @@ TODO: this should not use 0,0. It is wrong if
        # there's a pixmap. We should somehow compute the right
        # coordinates.
        pix.fill(self, QPoint(0, pixmap_blitting_y)) 
        # END
        
        currentY = 0
        i = self.parentList.scrollingOffset
        ht = self.height()
        while i < len(self.parentList.items) and currentY < ht:
            item = self.parentList.items[i]
            if i in items_to_draw:
                item.draw(painter, currentY
                          , pixmap_blitting_y, self.style()
                          , self.colorGroup(),self.palette()
                          , mousePos)
            currentY += item.height
            i += 1
        painter.end() # stop painting on pix
        painter.begin(self) # begin painting on the listarea
        painter.drawPixmap(0,pixmap_blitting_y,pix)
        painter.end()
        #print "end paint"
    def resizeEvent(self, ev):
        #print "resizeEvent : " + str(datetime.datetime.now())
        # no drawing must be done here! see qt docs for resizeEvent
        for item in self.parentList.items:
            item.widthChanged(ev.size().width())
        #print "centering item : " + str(datetime.datetime.now())
        if not (self.parentList.centeringState is None):
            self.parentList.centerItem(self.parentList.centeringState.item,
                self.parentList.centeringState.yoffset)
        #print "updateScrollButtonArrows : " + str(datetime.datetime.now())
        # BEGIN don't draw here!
        #self.paint(self.parentList.lastMousePos
        #           , invalidated_rect = ev.rect())
        # END
        self.parentList.updateScrollButtonArrows()
        #print "end resizeEvent " + str(datetime.datetime.now())

    def mouseMoveEvent(self, ev):
        # BEGIN debug
        #if self.parentList.lastMousePos:
        # print "mouseMoveEvent. lastpos = " + str(self.parentList.lastMousePos.y) + ", curpos = " + str(ev.pos().y())
        #else:
        #    print "mouseMoveEvent: lastMousePos = none"
        # END
        currentPosx = ev.pos().x()
        currentPosy = ev.pos().y()
        current_pos = MyPoint(currentPosx, currentPosy)
        if self.parentList.lastMousePos != None:
            ys = [ self.parentList.lastMousePos.y, ev.pos().y()]
            ys.sort()
            self.paint(current_pos , ys[0], ys[1])
        else:
            self.paint(current_pos , None, None)
        self.parentList.lastMousePos = current_pos
        
    def leaveEvent(self, ev):
        #print "leaveEvent. setting new pos = ", None
        self.parentList.lastMousePos = None
        self.paint(None, None, None)
        
    def mousePressEvent(self, ev):
        curY = 0
        self.lastMousePos = MyPoint(ev.pos().x(), ev.pos().y())
        for item in self.parentList.items[self.parentList.scrollingOffset:]:
            item.ifYouHaveBeenClickedEmitAppropriateSignal(mousePos = self.lastMousePos,
                                                           startY = curY,
                                                           signalEmitter = self.parentList
                                                           , pressed = True)
            curY += item.height

    def mouseReleaseEvent(self, ev):
        curY = 0
        self.lastMousePos = MyPoint(ev.pos().x(), ev.pos().y())
        for item in self.parentList.items[self.parentList.scrollingOffset:]:
            item.ifYouHaveBeenClickedEmitAppropriateSignal(mousePos = self.lastMousePos,
                                                           startY = curY,
                                                           signalEmitter = self.parentList
                                                           , pressed = False)
            curY += item.height

            
class TextSeparator:
    def __init__(self, text, width):
        self.text = text
        self.widthChanged(width)
    def ifYouHaveBeenClickedEmitAppropriateSignal(self, mousePos, startY, signalEmitter, pressed):
        pass
    def widthChanged(self, wt):
        self.width = wt
        # BEGIN calculate height and textHeight
        f = QFont()
        qfontm = QFontMetrics(f)
        boundingRectangle = qfontm.boundingRect(0,0,wt,1000,Qt.BreakAnywhere, self.text)
        self.height = boundingRectangle.height() + config.gSpaceAboveTextSeparator + config.gSpaceBelowTextSeparator
        self.textHeight = boundingRectangle.height()
        # END
    # TextSeparator
    def draw(self, painter, my_y_in_list
             , y_of_pixmap_in_list, style
             , colorGroup, palette, mouse):
        painter.setPen(palette.active().link())
        # BEGIN debug
        #print "drawing textseparator at y = " +  str(my_y_in_list - y_of_pixmap_in_list) + "in pixmap"
        #r2 = QRect(0, 0, self.width, self.height)
        #painter.fillRect(r2, QBrush(palette.active().dark()));
        # END
        painter.drawText(0 
                         ,  my_y_in_list - y_of_pixmap_in_list
                         + config.gSpaceAboveTextSeparator
                         , self.width, self.height
                         , Qt.BreakAnywhere, self.text)

class EmptySeparator:
    def __init__(self, width):
        self.widthChanged(width)
    def ifYouHaveBeenClickedEmitAppropriateSignal(self, mousePos, startY, signalEmitter, pressed):
        pass
    def widthChanged(self, wt):
        self.width = wt
        self.height = 12
    # EmptySeparator
    def draw(self, painter, my_y_in_list
             , y_of_pixmap_in_list, style
             , colorGroup, palette, mouse):
        pass


class text_item_button_info:
    def __init__(self, pixmaps, meaning, pushbutton_info):
        self.pixmaps = pixmaps
        self.meaning = meaning
        self.pushbutton_info = pushbutton_info
        
class TextItem:
    def __init__(self, text, icon, buttons_info, isSelectable, pressed, data
                 , parent_list
                 , resolution_mult):
        self.text  = text
        self.resolution_mult = resolution_mult
        assert isinstance(icon, QPixmap)
        self.icon = icon
        self.buttons_info = buttons_info
        self.isSelectable = isSelectable
        self.pressed = pressed
        self.parent_list = parent_list
        self.widthChanged(parent_list.listArea.width())
        self.data = data
        self.button_pushed = {} # map int -> bool
        for i in range(0, len(self.buttons_info)):
            inf = self.buttons_info[i]
            if cmp(inf.pushbutton_info, "not a togglebutton") == 0:
                pass
            elif cmp(inf.pushbutton_info , "pressed") == 0:
                self.button_pushed[i] = True
            else:
                self.button_pushed[i] = False
                
    def widthChanged(self, w):
        self.width = w
        # BEGIN calculate self.total_buttons_wt
        self.total_buttons_wt = int(config.g_text_item_button_width * self.resolution_mult) * len(self.buttons_info)
        # END
        
        # BEGIN calculate availableWtForText
        availableWtForText = w - config.gItemIconWt - self.total_buttons_wt
        if availableWtForText < 1:
            availableWtForText = 1
        # END
        
        # BEGIN calculate self.height and self.YOffsetForDrawingText
        f = QFont()
        qfontm = QFontMetrics(f)
        boundingRectangle = qfontm.boundingRect(0,0, availableWtForText,
            1000, Qt.BreakAnywhere, self.text)
        if  boundingRectangle.height() < config.gMinItemHt:
            self.height = config.gMinItemHt
            self.YOffsetForDrawingText = (config.gMinItemHt - boundingRectangle.height())/2
        else:
            val = int(4.0 * self.resolution_mult)
            self.height= boundingRectangle.height() + val * 2  
            self.YOffsetForDrawingText = val
        # END
    def ifYouHaveBeenClickedEmitAppropriateSignal(self, mousePos, startY, signalEmitter, pressed):
        if pressed:
            if self.isSelectable and self.thePointIsInsideMyText(point = mousePos, myStartY = startY):
                signalEmitter.emit(PYSIGNAL("item_pressed"), (signalEmitter, self))
            res = self.button_icon_the_pointer_is_inside(point = mousePos, myStartY = startY)
            if res != None:
                (i,meaning) = res
                if cmp(self.buttons_info[i].pushbutton_info, "not a togglebutton") != 0:
                    if self.button_pushed.get(i, False) == False:
                        self.button_pushed[i] = True
                        signalEmitter.emit(PYSIGNAL("item_togglebutton_pressed"), (signalEmitter, self, meaning))
                    else:
                        self.button_pushed[i] = False
                        signalEmitter.emit(PYSIGNAL("item_togglebutton_released"), (signalEmitter, self, meaning))
                else:
                    signalEmitter.emit(PYSIGNAL("item_button_pressed"), (signalEmitter, self, meaning))
        else:
            if self.isSelectable and self.thePointIsInsideMyText(point = mousePos, myStartY = startY):
                signalEmitter.emit(PYSIGNAL("item_released"), (signalEmitter, self))
            res = self.button_icon_the_pointer_is_inside(point = mousePos, myStartY = startY)
            if res != None:
                (i,meaning) = res
                if self.buttons_info[i].pushbutton_info == "not a togglebutton":
                    signalEmitter.emit(PYSIGNAL("item_button_released")
                                       , (signalEmitter, self, meaning))

    def thePointIsInsideMyText(self, point, myStartY):
        textRightBound = self.width - self.total_buttons_wt
        return ( point.x < textRightBound
                 and myStartY <= point.y and point.y < myStartY+ self.height)

    def button_icon_the_pointer_is_inside(self, point, myStartY):
        textRightBound = self.width - self.total_buttons_wt
        for i in range(0, len(self.buttons_info)):
            left_bound_icon_i = textRightBound + i * int(config.g_text_item_button_width * self.resolution_mult)
            right_bound_icon_i = left_bound_icon_i + int(config.g_text_item_button_width * self.resolution_mult)
            if ( point.x >= left_bound_icon_i and point.x < right_bound_icon_i
                 and myStartY<= point.y and point.y < myStartY+self.height):
                but_info = self.buttons_info[i]
                return (i, but_info.meaning)
        return None
    # TextItem.draw
    def draw(self, painter, y, pixmap_blitting_y
             ,style, colorGroup, palette, mouse):
        blit_y = y - pixmap_blitting_y
        textRightBound = self.width - self.total_buttons_wt
        # BEGIN draw the mouseover highlight around text and buttons
        if not self.pressed and exists( range(0, len(self.buttons_info))
                   , lambda i: self.button_pushed.get(i, False)):
            r2 = QRect(0, blit_y, textRightBound, self.height)

            highl = palette.active().highlight()
            ligh =  palette.active().light()
            
            bg = 0.5
            text = 0.5
            red = highl.red()*text + ligh.red()*bg
            green = highl.green()*text + ligh.green()*bg
            blue = highl.blue()*text + ligh.blue()*bg

            
            painter.fillRect(r2, QBrush(QColor(red, green, blue)))

        if self.pressed:
            r = QRect(0, y - pixmap_blitting_y, textRightBound, self.height)
            style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                                QStyle.Style_Enabled | QStyle.Style_Down)
        elif ( self.isSelectable and not mouse is None
               and self.thePointIsInsideMyText(point = mouse, myStartY = y)):
            r = QRect(0, blit_y, textRightBound, self.height)
            style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                        QStyle.Style_Enabled )
            painter.setPen(Qt.SolidLine)
        if mouse != None:
            result = self.button_icon_the_pointer_is_inside(point = mouse, myStartY = y)
            if  result != None:
                (ic, meaning) = result
                r2 = QRect(0, blit_y, textRightBound, self.height)
                painter.fillRect(r2, QBrush(palette.active().light()));
                r = QRect(textRightBound + ic
                          * int(config.g_text_item_button_width * self.resolution_mult)
                          , blit_y
                          , int(config.g_text_item_button_width * self.resolution_mult)
                          , self.height)
                style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                            QStyle.Style_Enabled )
                painter.setPen(Qt.SolidLine)
        # END
        # BEGIN draw the pressed button
        for i in range(0, len(self.buttons_info)):
            if self.button_pushed.get(i, False):
                r = QRect(textRightBound + i * int(config.g_text_item_button_width
                                                   * self.resolution_mult)
                          , blit_y
                          , int(config.g_text_item_button_width * self.resolution_mult)
                          , self.height)
                style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                                    QStyle.Style_Enabled | QStyle.Style_Down)

        # END
        
        # BEGIN draw text
        if self.isSelectable:
            if self.parent_list.text_color_for_items != None:
                painter.setPen(self.parent_list.text_color_for_items)
            else:
                painter.setPen(palette.active().text())
        else: # @@ flashing is ignored for selected items! Fix. TODO.
            bg = 0.4
            text = 0.6
            red = palette.active().text().red()*text + palette.active().background().red()*bg
            green = palette.active().text().green()*text + palette.active().background().green()*bg
            blue = palette.active().text().blue()*text + palette.active().background().blue()*bg
            painter.setPen(QColor(red, green, blue))
        painter.drawText(config.gItemIconWt, blit_y+self.YOffsetForDrawingText,
                         self.width - config.gItemIconWt - self.total_buttons_wt,
                         self.height, Qt.BreakAnywhere, self.text)
        # END

        # BEGIN draw button icons
        for i in range(0,len(self.buttons_info)):
            # BEGIN find which pixmap to use
            inf = self.buttons_info[i]
            if inf.pushbutton_info != "not a togglebutton" and self.button_pushed[i]:
                pix = inf.pixmaps[1]
            else:
                if mouse != None:
                    result = self.button_icon_the_pointer_is_inside(point = mouse, myStartY = y)
                    if  result != None:
                        (index, meaning) = result
                        if index == i:
                            pix = inf.pixmaps[1]
                        else:
                            pix = inf.pixmaps[0]
                    else:
                        pix = inf.pixmaps[0]
                else:
                    pix = inf.pixmaps[0]
            # END

            assert self.height > pix.height()
            if pix.width() >  int(config.g_text_item_button_width * self.resolution_mult):
                pix.resize( int(config.g_text_item_button_width * self.resolution_mult)
                            , pix.height())
            gap_x =  int(config.g_text_item_button_width
                         * self.resolution_mult) - pix.width()
            gap_y =  self.height - pix.height()
            painter.drawPixmap(textRightBound + i
                               * int(config.g_text_item_button_width * self.resolution_mult)
                               + gap_x/2
                               , blit_y + gap_y/2, pix)
        # END
            
        # BEGIN draw leftmost icon
        if self.icon.height() > self.height:
            iconYOffset = 0
        else:
            iconYOffset = (self.height - self.icon.height())/2
        if self.icon.width() > config.gItemIconWt:
            iconXOffset = 0
        else:
            iconXOffset = (config.gItemIconWt - self.icon.width())/2
        painter.drawPixmap(QPoint(iconXOffset, blit_y + iconYOffset), self.icon)
        # END

class CenteringCommand:
    def __init__(self, box, itemText, yoffset):
        self.itemText = itemText
        self.box = box
        self.yoffset = yoffset
    
class CenteringState:
    def __init__(self, item, yoffset):
        self.item = item
        self.yoffset = yoffset


class FlashingStateYesAnimating:
    def __init__(self, remainingFlashesBeforeStop, resetOriginalColorsWhenFinished):
        self.resetOriginalColorsWhenFinished = resetOriginalColorsWhenFinished
        self.remainingFlashesBeforeStop = remainingFlashesBeforeStop
class FlashingStateYesStill:
    def __init__(self):
        pass
class FlashingStateNo:
    def __init__(self):
        pass
        
class MyListBox(QHBox):
    def enable_all_items(self, selectable):
        for i in self.items:
            if isinstance(i, TextItem):
                i.isSelectable = selectable
    def set_pushbutton_of_textitem(self
                                   # identifies the button
                                   , button_meaning
                                   , pressed
                                   # identifies the item
                                   , item_data):
        it = find(self.items
                  , lambda i: isinstance(i, TextItem) and i.data == item_data)
        if it != None:
            for b in range(0, len(it.buttons_info)):
                if (  it.buttons_info[b].pushbutton_info != "not a togglebutton"
                      and it.buttons_info[b].meaning == button_meaning):
                    it.button_pushed[b] = pressed
    def enable_these_items(self, items):
        for i in self.items:
            if isinstance(i, TextItem):
                if i.data in items:
                    i.isSelectable = True
                else:
                    i.isSelectable = False

    def isFlashing(self):
        return not isinstance(self.flashingState, FlashingStateNo)

    def flash_all_items(self, howManyFlashes):
        self.stopFlashing()
        self.flashingState = FlashingStateYesAnimating(howManyFlashes
                                                       , "the reset flag is useless here")
        self.startTimer(90)

    def stopFlashing(self):
        self.killTimers()
        self.flashingState = FlashingStateNo()
        self.text_color_for_items = None
        self.listArea.paint( mousePos = None
                             , invalidated_y_bottom = None
                             , invalidated_y_top = None)
        
    # MyListBox.timerEvent
    def timerEvent(self, ev):
        #print "enter timerEvent. mousepos = ", point_to_string(self.lastMousePos)

        # BEGIN compute new text color for the items
        col = QColor() #(63, 143, 91)
        col.setHsv( (self.flashingState.remainingFlashesBeforeStop*10) % 255
                    , 133
                    ,(self.flashingState.remainingFlashesBeforeStop*70) % 255)
        self.text_color_for_items = col
        # END
        

        self.listArea.paint(mousePos = None
                            , invalidated_y_bottom = None
                            , invalidated_y_top = None)
        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing()
        #print "exit timerEvent. mousepos = ", point_to_string(self.lastMousePos)
    def __init__(self, frameVisible, parent, resolution_mult
                 , up_pixm, down_pixm, top_pixm, bottom_pixm):
        QHBox.__init__(self, parent)
        self.up_pixm = up_pixm
        self.down_pixm = down_pixm
        self.top_pixm = top_pixm
        self.bottom_pixm = bottom_pixm
        self.items = []
        self.centeringState = None
        self.frameVisible = frameVisible
        self.text_color_for_items = None
        # BEGIN create ListArea
        self.listArea = ListArea(parentList = self
                                 , parentWidget = self) #self.vbox)
        self.listArea.show()
        self.listArea.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Ignored)
        # END

        wt = int (config.width_of_scroll_buttons * resolution_mult)
        ht = int (config.width_of_scroll_buttons * resolution_mult)
        
        self.upAndDownButtonBox = QVBox(self)
        self.upAndDownButtonBox.show()
        self.upAndDownButtonBox.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.upAndDownButtonBox.setMaximumWidth(wt)
        
        # BEGIN up button box
        self.buttonBoxUp = QVBox(self.upAndDownButtonBox)
        self.buttonBoxUp.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.buttonBoxUp.show()

        self.gotoTopBut = QPushButton(self.buttonBoxUp)
        self.gotoTopBut.show()
        self.gotoTopBut.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.gotoTopBut.setMaximumHeight(ht)
        QObject.connect(self.gotoTopBut, SIGNAL("clicked()"), self.onGotoTopClicked)

        self.buttonUp = QPushButton(self.buttonBoxUp)
        self.buttonUp.show()
        self.buttonUp.setMinimumWidth(wt)
        self.buttonUp.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        QObject.connect(self.buttonUp, SIGNAL("clicked()"), self.onPageUpClicked)
        # END

        # BEGIN down button box
        self.buttonBoxDown = QVBox(self.upAndDownButtonBox)
        self.buttonBoxDown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.buttonBoxDown.show()

        self.buttonDown = QPushButton(self.buttonBoxDown)
        self.buttonDown.show()
        self.buttonDown.setMinimumWidth(wt)
        self.buttonDown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        QObject.connect(self.buttonDown, SIGNAL("clicked()"), self.onPageDownClicked)
        
        self.gotoBottomBut = QPushButton(self.buttonBoxDown)
        self.gotoBottomBut.show()
        self.gotoBottomBut.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.gotoBottomBut.setMaximumHeight(ht)
        QObject.connect(self.gotoBottomBut, SIGNAL("clicked()"), self.onGotoBottomClicked)
        # END

        # BEGIN set pixmaps for scrool buttons
        self.buttonUp.setPixmap(QPixmap())
        self.buttonDown.setPixmap(QPixmap())
        self.gotoBottomBut.setPixmap(QPixmap())
        self.gotoTopBut.setPixmap(QPixmap())

#        self.buttonUp.setPixmap(QPixmap(config.imgDir + "stock_up.png"))
#        self.gotoTopBut.setPixmap(QPixmap(config.imgDir + "stock_top.png"))
#        self.buttonDown.setPixmap(QPixmap(config.imgDir + "stock_down.png"))
#        self.gotoBottomBut.setPixmap(QPixmap(config.imgDir + "stock_bottom.png"))

        # END
        self.setSpacing(int (5.0 * resolution_mult))
        self.scrollingOffset = 0
        self.lastMousePos = None
        
        
    def clear(self):
        self.items = []
        self.scrollingOffset = 0
        self.centeringState = None
        #print "clear. mousepos = ", point_to_string(self.lastMousePos)
        
    def onLineUpClicked(self):
        self.centeringState = None
        self.scrollUpLineWithoutUpdating()
        self.make_changes_visible()

    def scrollUpLineWithoutUpdating(self):
        if self.scrollingOffset != 0:
            self.scrollingOffset -= 1

    # BEGIN itemNPixelsAboveTheItem
    #
    # This function is a bit complicated, and deserves explanation.
    #
    #    itemNPixelsAboveTheItem(self, n, i) = j
    #
    # means
    #
    # "j is the maximum index of an item J such that: the distance
    # from J to I is greater than n. (where I is such that index(I) =
    # i)."
    #
    # We want to write this in a form that is as declarative as
    # possible, i.e. as close as possible to the above definition. To
    # do that, we first define the function distance(j, i):
    #
    def __itemNPixelsAboveTheItem(self, n, i):
        def distance(j, i):
            assert j <= i
            return reduce(operator.add,
                          [ item.height for item in self.items[j:i] ],
                          0)
        # now we apply the definition of j: j is the maximum of the
        # set of all indices in range(0,i) such that...
        distant_elements =  [ j for j in range(0,i) if  distance(j, i) >= n]
        if distant_elements ==[]:
            return None
        else:
            return max(distant_elements)

    # This is an imperative version by Thomas. --- update: Sorry
    # Thomas, this does not work: try scrolling page up to
    # verify. When I have time I'll try to see why. In the meantime,
    # let us use the declarative version. Unfortunately I had already
    # thrown away my old version when I saw yours.
    #
    # A: I accidently wrote ItemNPi... (big I), so it didnt work because
    # of that typo.
    def itemNPixelsAboveTheItem(self, n, i):
        deltaY = 0
        while i != 0:
            i -= 1
            deltaY += self.items[i].height
            if deltaY > n:
                return i + 1
        return None
    # END itemNPixelsAboveTheItem        

    def onPageUpClicked(self):
        self.centeringState = None
        self.scrollUpPageWithoutUpdating()
        self.make_changes_visible()
    def scrollDownLineWithoutUpdating(self):
        if self.scrollingOffset != len(self.items)-1:
            self.scrollingOffset += 1
    def scrollUpPageWithoutUpdating(self):
        if self.scrollingOffset > 0:
            i = self.itemNPixelsAboveTheItem(self.listArea.height()-1, self.scrollingOffset)
            if i is None:
                self.scrollingOffset = 0
            else:
                self.scrollingOffset = i
    def onLineDownClicked(self):
        self.centeringState = None
        self.scrollDownLineWithoutUpdating()
        self.make_changes_visible()
    def onPageDownClicked(self):
        self.centeringState = None
        x = self.itemAtPosition(self.listArea.height()-1)
        if not x is None:
            self.scrollingOffset = x
            self.listArea.paint(mousePos = None
                                , invalidated_y_bottom = None
                                , invalidated_y_top = None)
        self.updateScrollButtonArrows()
    def onGotoTopClicked(self):
        self.centeringState = None
        if self.scrollingOffset != 0:
            self.scrollingOffset = 0
            self.make_changes_visible()
    def onGotoBottomClicked(self):
        self.centeringState = None
        self.scrollingOffset = len(self.items)-1
        self.scrollUpPageWithoutUpdating()
        self.scrollDownLineWithoutUpdating()
        x = self.itemAtPosition(self.listArea.height()-1)
        while x != None and self.scrollingOffset != len(self.items)-1:
            self.scrollingOffset += 1
            x = self.itemAtPosition(self.listArea.height()-1)
        self.make_changes_visible()
    def updateScrollButtonArrows(self):
        if self.theUpScrollArrowMustBeVisible():
            if self.buttonUp.pixmap().isNull():
                self.buttonUp.setPixmap(self.up_pixm)
            if self.gotoTopBut.pixmap().isNull():
                self.gotoTopBut.setPixmap(self.top_pixm)
            if not self.buttonBoxUp.isEnabled():
                self.buttonBoxUp.setEnabled(True)
            if not self.gotoTopBut.isEnabled():
                self.gotoTopBut.setEnabled(True)
        else:
            if not self.buttonUp.pixmap().isNull():
                self.buttonUp.setPixmap(QPixmap())
            if not self.gotoTopBut.pixmap().isNull():
                self.gotoTopBut.setPixmap(QPixmap())
            if self.buttonBoxUp.isEnabled():
                self.buttonBoxUp.setEnabled(False)
            if self.gotoTopBut.isEnabled():
                self.gotoTopBut.setEnabled(False)
        if self.theDownScrollArrowMustBeVisible():
            if self.buttonDown.pixmap().isNull():
                self.buttonDown.setPixmap(self.down_pixm)
            if self.gotoBottomBut.pixmap().isNull():
                self.gotoBottomBut.setPixmap(self.bottom_pixm)
            if not self.buttonBoxDown.isEnabled():
                self.buttonBoxDown.setEnabled(True)
            if not self.gotoBottomBut.isEnabled():
                self.gotoBottomBut.setEnabled(True)
        else:
            if not self.buttonDown.pixmap().isNull():
                self.buttonDown.setPixmap(QPixmap())
            if not self.gotoBottomBut.pixmap().isNull():
                self.gotoBottomBut.setPixmap(QPixmap())
            if self.buttonBoxDown.isEnabled():
                self.buttonBoxDown.setEnabled(False)

            if self.gotoBottomBut.isEnabled():
                self.gotoBottomBut.setEnabled(False)
    def theUpScrollArrowMustBeVisible(self):
        return len(self.items) > 0 and self.scrollingOffset > 0
    def theDownScrollArrowMustBeVisible(self):
        # @@ FIXME: bug if the last element of the list is
        # partially visible: you see half of it, and yet you
        # cannot scroll.
        return len(self.items) > 0 and  self.itemAtPosition(self.listArea.height()-1) != None
    #self.lastVisibleItem() < len(self.items)-1
      
    def lastVisibleItem(self):
        if len(self.items) > 0:
            x = self.itemAtPosition(self.listArea.height()-1)
            if x == None:
                return len(self.items) -1
            else:
                return x
        else:
            return None

    # this is more declarative (and slower) than the following. Feel
    # free to comment this and uncomment the imperative version.
##     def itemAtPosition(self, posy):
##         def distance(i, j, delta):
##             d1 = reduce( operator.add,
##                          [ item.height for item in self.items[i:j-1] ],
##                          0)
##             d2 = d1 + self.items[j].height
##             return delta in range(d1, d2)
        
##         return find( range(self.scrollingOffset, len(self.items)),
##                      lambda i: distance(self.scrollingOffset, i, posy))

    # the item at a given position can be None
    def itemAtPosition(self, posy):
        start_y_of_cur_item = 0
        i = self.scrollingOffset
        while i < len(self.items) and start_y_of_cur_item < self.listArea.height():
            cur_item = self.items[i]
            # if item.height == 4 and posy = 3, I must return the
            # item. But if posy == item.height == 4, I must return the
            # following item. So the equal sign is not allowed:
            if posy < start_y_of_cur_item + cur_item.height :
                return i
            else:
                start_y_of_cur_item += cur_item.height
                i += 1
        return None

    # call this after adding items, or you won't see the changes
    def make_changes_visible(self):
        #print "make_changes_visible called. mousePos = " , self.lastMousePos
        self.listArea.paint(self.lastMousePos
                            , invalidated_y_top = None
                            , invalidated_y_bottom = None)
        self.updateScrollButtonArrows()
    def yOffsetOfItem(self, item):
        # return the sum of the heights of all the items from
        # scrollingOffset to item :-)
        return reduce(operator.add ,
                      [i.height for i in self.items[self.scrollingOffset:self.items.index(item)]],
                      0)
    # changes self.scrollingOffset in order to show item in position
    # yoffset from the top (if possible)
    def centerItem(self, itemText, yoffset):
        #print "centering item  ", itemText, ' to y = ', yoffset
        item = find(self.items, lambda i:i.text == itemText)
        assert item != None
        self.centeringState = CenteringState(itemText, yoffset)
        # BEGIN calculate realYOffset for the item
        theItemWouldBeCompletelyInside = yoffset + item.height < self.listArea.height()
        if theItemWouldBeCompletelyInside:
            realYOffset = yoffset
        else:
            realYOffset = (self.listArea.height()-item.height) * 3/4
        # END

        # BEGIN calculate self.scrollingOffset

        # BEGIN mathematical definition (slow)
##         itemIdx = self.items.index(item)
##         i = find( range(itemIdx),
##                   lambda i: realYOffset >= sum( [x.height for x in self.items[i: itemIdx ]], 0))
##         if i != None:
##             self.scrollingOffset = i
##         else:
##             self.scrollingOffset = 0
        #END

        # BEGIN imperative version (fast)
        currentOffsetY = realYOffset
        i = self.items.index(item)
        while currentOffsetY > 0:
            if i > 0:
                i -= 1
                currentOffsetY -= self.items[i].height
            else:
                self.scrollingOffset = 0
                return
        self.scrollingOffset = i
        # END
        
        # END 


class WrappingButton(QHBox):
    def __init__(self, text, parent, data, toggle):
        QHBox.__init__(self, parent)
        self.text = text
        self.toggle_button = toggle
        self.pressed = False
        self.mouseOver = False
        self.setMargin(5)
        #self.setBackgroundMode(Qt.NoBackground)
        self.recalculateHeight(self.width())
        self.data = data
    def setText(self, t):
        self.text = t
        self.recalculateHeight(self.width())
        self.repaint()
    def resizeEvent(self, ev):
        self.recalculateHeight(ev.size().width())
        self.repaint()
    def setPressedState(self, pressed):
        self.pressed = pressed
        self.repaint()
    def recalculateHeight(self, wt):
        f = QFont()
        qfontm = QFontMetrics(f)
        boundingRectangle = qfontm.boundingRect(0,0,
            wt - config.gWrappingButton_DistanceOfTextFromBorder*2,
            1000,
            Qt.BreakAnywhere,
            self.text)
        newHt = boundingRectangle.height() + config.gWrappingButton_DistanceOfTextFromBorder*2
        if newHt != self.height():
            self.setMinimumHeight(newHt)
    def mousePressEvent(self, mouseEvent):
        if self.toggle_button:
            self.pressed = not self.pressed
            if self.pressed:
                self.emit(PYSIGNAL("pressed"), (self,))
            else:
                self.emit(PYSIGNAL("released"), (self,))
        else:
                self.emit(PYSIGNAL("pressed"), (self,))
        self.repaint()
    def enterEvent(self, qevent):
        self.mouseOver = True
        self.repaint()
    def leaveEvent(self, qev):
        self.mouseOver = False
        self.repaint()
    def paintEvent(self, pevent):
        pix = QPixmap(self.size())
        painter = QPainter(pix)
        # BEGIN draw the button
        pix.fill(self, QPoint(0,0)) # uses the background pixmap if
                                    # any
        r = QRect(0,0,self.width(), self.height())
        if self.isEnabled():
            lStyle = QStyle.Style_Enabled 
            if self.pressed:
                lStyle |= QStyle.Style_Down
            if self.mouseOver:
                lStyle |= QStyle.Style_MouseOver
            self.style().drawPrimitive(QStyle.PE_ButtonCommand,
                           painter,
                           r,
                           self.colorGroup(),
                           lStyle)
        # END
        
        # BEGIN draw the text
        painter.setPen(self.palette().active().text())
        painter.drawText(config.gWrappingButton_DistanceOfTextFromBorder
                         , config.gWrappingButton_DistanceOfTextFromBorder,
            self.width() - config.gWrappingButton_DistanceOfTextFromBorder*2,
            self.height() - config.gWrappingButton_DistanceOfTextFromBorder,
            Qt.BreakAnywhere, self.text)
        # END
        painter.end() # it is ok to do end before begin
        painter.begin(self)
        painter.drawPixmap(0,0,pix)


class MyLineEdit(QLineEdit):
    def __init__(self, parent):
        QLineEdit.__init__(self, parent)
        self.timerId = None
    def mousePressEvent(self, e):
        QLineEdit.selectAll(self)
    def timerEvent(self, ev):
        if ev.timerId() == self.timerId:
            self.emit(PYSIGNAL("timer"), (self,))
            # do not use killTimers here! QLineEdit uses its own timer
            # to manage the blinking cursor!
            self.killTimer(self.timerId)
            self.timerId = None
    def keyPressEvent(self, e):
        if self.timerId != None:
            # do not use killTimers here! QLineEdit uses its own timer
            # to manage the blinking cursor!
            self.killTimer(self.timerId)
        self.timerId = self.startTimer(1000)
        e.ignore()
        QLineEdit.keyPressEvent(self, e)

        
class MyComboBox(QComboBox):
    def __init__(self, parent):
        QComboBox.__init__(self, parent)
        QObject.connect(self, SIGNAL("activated(int)"), self.onActivated)
        self.currentId = 0
        self.idToItem = {}
    def insertItem(self,  icon, text, data):
        QComboBox.insertItem(self, QPixmap(icon), text, self.currentId)
        self.idToItem[self.currentId] = data
        self.currentId += 1
    def onActivated(self, i):
        self.emit(PYSIGNAL("itemActivated"), (self, self.idToItem[i])) 
    def clear(self):
        self.idToItem = {}
        self.currentId  = 0
        QPopupMenu.clear(self)

class PopupMenuItem:
    def __init__(self, name):
        self.name = name
    def stringForPopupMenu(self):
        return self.name

class MyPopupMenu(QPopupMenu):
    def __init__(self, parent):
        QPopupMenu.__init__(self, parent)
        self.currentId = 0
        self.fromIdToItem = {}
        QObject.connect(self, SIGNAL("activated(int)"), self.onActivated)
    def addItem(self, item, iconset, get_string):
        self.insertItem(iconset, get_string(item), self.currentId)
        self.fromIdToItem[self.currentId] = item
        self.currentId += 1
    def onActivated(self, id):
        item = self.fromIdToItem[id]
        self.emit(PYSIGNAL("itemActivated"), (self, item))
    def clear(self):
        self.fromIdToItem = {}
        self.currentId  = 0
        QPopupMenu.clear(self)

